/*
 * Copyright 2002-2022 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.springframework.security.web.server.csrf;

import reactor.core.publisher.Mono;

import org.springframework.web.server.ServerWebExchange;

/**
 * Implementations of this interface are capable of resolving the token value of a
 * {@link CsrfToken} from the provided {@code ServerWebExchange}. Used by the
 * {@link CsrfWebFilter}.
 *
 * @author Steve Riesenberg
 * @since 5.8
 * @see ServerCsrfTokenRequestAttributeHandler
 */
@FunctionalInterface
public interface ServerCsrfTokenRequestResolver {

	/**
	 * Returns the token value resolved from the provided {@code ServerWebExchange} and
	 * {@link CsrfToken} or {@code Mono.empty()} if not available.
	 * @param exchange the {@code ServerWebExchange} with the request being processed
	 * @param csrfToken the {@link CsrfToken} created by the
	 * {@link ServerCsrfTokenRepository}
	 * @return the token value resolved from the request
	 */
	Mono<String> resolveCsrfTokenValue(ServerWebExchange exchange, CsrfToken csrfToken);

}
